from janis_core.utils.errors import (
    TooManyArgsException,
    IncorrectArgsException,
    InvalidByProductException,
    ConflictingArgumentsException,
)
from janis_core.utils.logger import Logger

from janis_core.utils.bracketmatching import get_keywords_between_braces
from abc import ABC


class Selector(ABC):
    pass


class InputSelector(Selector):
    def __init__(self, input_to_select, use_basename=None):
        # maybe worth validating the input_to_select identifier
        self.input_to_select = input_to_select
        self.use_basename = use_basename

    def to_string_formatter(self):
        kwarg = {self.input_to_select: self}
        return StringFormatter(f"{{{self.input_to_select}}}", **kwarg)

    def __radd__(self, other):
        return StringFormatter(other) + self.to_string_formatter()

    def __add__(self, other):
        return self.to_string_formatter() + other


class WildcardSelector(Selector):
    def __init__(self, wildcard, first_element=None):
        self.wildcard = wildcard
        self.first_element = first_element


class MemorySelector(InputSelector):
    def __init__(self):
        super().__init__("runtime_memory")


class CpuSelector(InputSelector):
    def __init__(self, default=1):
        super().__init__("runtime_cpu")
        self.default = default


class StringFormatter(Selector):
    def __init__(self, format: str, **kwargs):
        self._format: str = format

        keywords, balance = get_keywords_between_braces(self._format)

        if balance > 0:
            Logger.warn(
                "There was an imbalance of braces in the string _format, this might cause issues with concatenation"
            )

        skwargs = set(kwargs.keys())

        if not keywords == skwargs:
            # what's the differences
            if not keywords.issubset(skwargs):
                raise IncorrectArgsException(
                    "The _format required additional arguments to be provided by "
                    "**kwargs, requires the keys:" + ", ".join(keywords - skwargs)
                )
            else:
                raise TooManyArgsException(
                    "The **kwargs contained unrecognised keys: "
                    + ", ".join(skwargs - keywords)
                )

        self.kwargs = kwargs

    resolved_types = [str, int, float]

    def resolve_with_resolved_values(self, **resolved_values):

        s1 = set(self.kwargs.keys())
        actual_keys, _ = get_keywords_between_braces(self._format)
        if s1 != actual_keys:
            diff = (actual_keys - s1).union(s1 - actual_keys)

            raise Exception(
                "The format for the string builder has changed since runtime, or an internal error has"
                " occurred. The following keys did not appear in both sets: "
                + ", ".join(diff)
            )

        s2 = set(resolved_values.keys())

        missing_keys = s1 - s2
        if len(missing_keys) > 0:
            raise IncorrectArgsException(
                "There were missing parameters when formatting string: "
                + ", ".join(missing_keys)
            )

        unresolved_values = [
            f"{r} ({type(resolved_values[r]).__name__})"
            for r in resolved_values
            if not any(
                isinstance(resolved_values[r], t)
                for t in StringFormatter.resolved_types
            )
        ]
        if len(unresolved_values) > 0:
            raise ValueError(
                "There were unresolved parameters when formatting string: "
                + ", ".join(unresolved_values)
            )

        retval = self._format
        for k in resolved_values:
            retval = retval.replace(f"{{{k}}}", str(resolved_values[k]))
        return retval

    def __radd__(self, other):
        return StringFormatter(other) + self

    def __add__(self, other):
        if isinstance(other, str):
            # check if it has args in it
            keywords = get_keywords_between_braces(other)
            if len(keywords) > 0:
                invalidkwargs = [k for k in self.kwargs if k not in self.kwargs]
                if len(invalidkwargs) > 0:
                    raise InvalidByProductException(
                        f"The string to be concatenated contained placeholder(s) ({', '.join(invalidkwargs)})"
                        f"that were not in the original StringFormatter"
                    )
            return self._create_new_formatter_from_strings_and_args(
                [self._format, other], **self.kwargs
            )

        elif isinstance(other, InputSelector):
            return self + other.to_string_formatter()

        elif isinstance(other, StringFormatter):
            # check if args overlap and they're different
            s1 = set(self.kwargs.keys())
            s2 = set(other.kwargs.keys())
            intersection = s1.intersection(s2)

            if len(intersection) > 0:
                not_same_args = [
                    k for k in intersection if self.kwargs[k] != other.kwargs[k]
                ]
                if len(not_same_args) > 0:
                    raise ConflictingArgumentsException(
                        f"Couldn't concatenate formats as there keys ({', '.join(not_same_args)}) "
                        f"that were not equal between formatters "
                    )

            # yeah we sweet
            new_args = {**self.kwargs, **other.kwargs}
            return StringFormatter._create_new_formatter_from_strings_and_args(
                [self._format, other._format], **new_args
            )

    @staticmethod
    def _create_new_formatter_from_strings_and_args(strings: [str], **kwargs):
        new_format = "".join(strings)
        try:
            return StringFormatter(new_format, **kwargs)
        except IncorrectArgsException as e:
            new_params = set(
                get_keywords_between_braces(new_format)[0] - set(kwargs.keys())
            )
            raise InvalidByProductException(
                "Joining the input files (to '{new_format}') created the new params: "
                + ", ".join(new_params)
            )



from typing import Dict, Any, List, Tuple


def first_value(d: Dict):
    return next(iter(d.values()))


def get_value_for_hints_and_ordered_resource_tuple(
    hints: Dict[str, Any], tuples: List[Tuple[str, Dict[str, int]]]
):
    if not hints:
        return None
    for k, d in tuples:
        if k not in hints:
            continue
        v = hints[k]
        if v not in d:
            continue
        return d[v]
    return None


def zip_directory(parent_dir, dir_name):
    import subprocess
    from .logger import Logger
    from os import chdir

    Logger.info("Zipping tools")
    chdir(parent_dir)

    zip_result = subprocess.run(["zip", "-r", f"{dir_name}.zip", f"{dir_name}/"])
    if zip_result.returncode == 0:
        Logger.info("Zipped tools")
    else:
        Logger.critical(zip_result.stderr)


def is_array_prefix(prefix, to):
    if len(prefix) > len(to):
        return False

    for i in range(len(prefix)):
        if prefix[i] != to[i]:
            return False

    return True


def recursive_2param_wrap(methodname, items):
    # It would be awesome if this is generic, but it opens up a new can of worms about how to wrap it
    # 2 might even be confusing, as you could say zip(zip(A, B), zip(C, D)) instead of zip(A, zip(B, zip(C, D)))
    # And more than length=2 you could have cases where the last function call might not have a clean 'n' params

    if len(items) < 1:
        raise Exception(
            "'recursive_2param_wrap' required two parameters to wrap a workflow"
        )
    if len(items) == 2:
        return f"{methodname}({items[0]}, {items[1]})"
    return f"{methodname}({items[0]}, {recursive_2param_wrap(methodname, items[1:])})"



from janis_core.utils.logger import Logger
from os import path as p, getcwd


class ExportPathKeywords:
    workflow_spec = "{language}"
    workflow_name = "{name}"

    default = "./{name}"  # f"~/Desktop/{workflow_name}/{workflow_spec}/"
    default_no_spec = "./{name}"  # f"~/Desktop/{workflow_name}/"

    @staticmethod
    def resolve(path, workflow_spec, workflow_name):

        if not path:
            Logger.critical("Output path was invalid, changed to working directory")
            path = "."

        if ExportPathKeywords.workflow_spec in path and workflow_spec is None:
            raise Exception(
                f"path ('{path}') contained parameter {ExportPathKeywords.workflow_spec} "
                "but caller of .resolve did not pass language"
            )

        if ExportPathKeywords.workflow_name in path and workflow_name is None:
            raise Exception(
                f"path ('{path}') contained parameter {ExportPathKeywords.workflow_name} "
                "but caller of .resolve did not pass workflow name"
            )

        path = p.expanduser(path)

        if (len(path) == 1 and path == ".") or path[:2] == "./":
            path = getcwd() + (path[1:] if len(path) > 1 else "")

        elif path[0] != "/":
            path = p.join(getcwd(), path)

        return path.replace(
            ExportPathKeywords.workflow_spec,
            workflow_spec.lower() if workflow_spec else "",
        ).replace(
            ExportPathKeywords.workflow_name,
            workflow_name.lower() if workflow_name else "",
        )



import re
from abc import ABC, abstractmethod
from typing import Optional, List, Dict, Any, Union

from janis_core.types import (
    ParseableType,
    Selector,
    Array,
    get_instantiated_type,
    DataType,
)
from janis_core.utils.logger import Logger
from janis_core.utils.metadata import Metadata
from janis_core.utils.validators import Validators

ToolType = str


class ToolTypes:
    Workflow: ToolType = "workflow"
    CommandTool: ToolType = "command-tool"
    ExpressionTool: ToolType = "expression-tool"


class TInput(object):
    def __init__(self, tag: str, intype: DataType, default=None):
        self.tag = tag
        self.intype = intype
        self.default = default

    def id(self):
        return self.tag


class TOutput(object):
    def __init__(self, tag, outtype):
        self.tag = tag
        self.outtype = outtype

    def id(self):
        return self.tag


class Tool(ABC, object):
    """
    One of Workflow, CommandLineTool, ExpressionTool* (* unimplemented)
    """

    def __init__(self, metadata_class=Metadata, **connections):
        """
        :param metadata_class:
        :param connections:
        """

        self.metadata: metadata_class = metadata_class()
        meta = self.bind_metadata()
        if meta:
            self.metadata = meta

        self.connections = connections

    @classmethod
    @abstractmethod
    def type(cls) -> ToolType:
        raise Exception(f"'{cls}' must implement type() method")

    @abstractmethod
    def id(self) -> str:
        raise Exception("Must implement id() method")

    @staticmethod
    def tool_module():
        return None

    @staticmethod
    def tool_provider():
        return None

    @abstractmethod
    def tool_inputs(self) -> List[TInput]:
        raise Exception("Must implement inputs() method")

    @abstractmethod
    def tool_outputs(self) -> List[TOutput]:
        raise Exception("Must implement outputs() method")

    def inputs_map(self) -> Dict[str, TInput]:
        return {inp.tag: inp for inp in self.tool_inputs()}

    def outputs_map(self) -> Dict[str, TOutput]:
        return {outp.tag: outp for outp in self.tool_outputs()}

    def friendly_name(self) -> Optional[str]:
        """
        Overriding this method is not required UNLESS you distribute your tool.
        Generating the docs will fail if your tool does not provide a name.

        :return: A friendly name of your tool
        """
        return None

    def all_input_keys(self) -> List[str]:
        return [t.id() for t in self.tool_inputs()]

    @abstractmethod
    def generate_inputs_override(
        self, additional_inputs=None, with_resource_overrides=False, hints=None
    ):
        pass

    def __call__(self, **connections):
        self.connections = connections
        return self

    @staticmethod
    @abstractmethod
    def version():
        return None

    def doc(self) -> Optional[str]:
        return None

    @abstractmethod
    def translate(
        self, translation: str, with_docker=True, with_resource_overrides=False
    ):
        raise Exception("Subclass must provide implementation for 'translate()' method")

    def bind_metadata(self):
        """
        A convenient place to add metadata about the tool. You are guaranteed that self.metadata will exist.
        It's possible to return a new instance of the ToolMetadata / WorkflowMetadata which will be rebound.
        This is usually called after the initialiser, though it may be called multiple times.
        :return:
        """
        return self.metadata



import re
from abc import ABC, abstractmethod
from typing import List, Dict, Optional, Any, Union

from janis_core.utils.validators import Validators

from janis_core.types import (
    Selector,
    Logger,
    ParseableType,
    get_instantiated_type,
    Stdout,
    Stderr,
)

from janis_core.types.common_data_types import String, Filename
from janis_core.tool.tool import Tool, ToolTypes, TInput, TOutput
from janis_core.enums.supportedtranslations import SupportedTranslation
from janis_core.utils.metadata import ToolMetadata, Metadata


class ToolArgument:
    expr_pattern = "\$\(.*\)"

    def __init__(
        self,
        value: Any,
        prefix: Optional[str] = None,
        position: Optional[int] = 0,
        separate_value_from_prefix=None,
        doc: Optional[str] = None,
        shell_quote: bool = None,
    ):
        """
        A ``ToolArgument`` is a CLI parameter that cannot be override (at runtime).
        The value can


        :param value:
        :type value: ``str`` | ``janis.InputSelector`` | ``janis.StringFormatter``
        :param position: The position of the input to be applied. (Default = 0, after the base_command).
        :param prefix: The prefix to be appended before the element. (By default, a space will also be applied, see ``separate_value_from_prefix`` for more information)
        :param separate_value_from_prefix: (Default: True) Add a space between the prefix and value when ``True``.
        :param doc: Documentation string for the argument, this is used to generate the tool documentation and provide
        :param shell_quote: Stops shell quotes from being applied in all circumstances, useful when joining multiple commands together.
        """

        self.prefix: Optional[str] = prefix
        self.value = value
        self.position: Optional[int] = position
        self.is_expression = (
            isinstance(self.value, Selector)
            or (re.match(self.expr_pattern, self.value) is not None)
            if self.value
            else None
        )
        self.separate_value_from_prefix = separate_value_from_prefix
        self.doc = doc
        self.shell_quote = shell_quote

        if (
            self.prefix
            and self.separate_value_from_prefix is not None
            and not self.separate_value_from_prefix
            and not self.prefix.endswith("=")
        ):
            # I don't really know what this means.
            Logger.warn(
                f"Argument ({self.prefix} {self.value}) is not separating and did not end with ='"
            )


# This should really be a CommandToolInput
class ToolInput(ToolArgument):
    def __init__(
        self,
        tag: str,
        input_type: ParseableType,
        position: Optional[int] = None,
        prefix: Optional[str] = None,
        separate_value_from_prefix: bool = None,
        prefix_applies_to_all_elements: bool = None,
        separator: str = None,
        shell_quote: bool = None,
        localise_file: bool = None,
        default: Any = None,
        doc: Optional[str] = None,
    ):
        """
        A ``ToolInput`` represents an input to a tool, with parameters that allow it to be bound on the command line.
        The ToolInput must have either a position or prefix set to be bound onto the command line.

        :param tag: The identifier of the input (unique to inputs and outputs of a tool)
        :param input_type: The data type that this input accepts
        :type input_type: ``janis.ParseableType``
        :param position: The position of the input to be applied. (Default = 0, after the base_command).
        :param prefix: The prefix to be appended before the element. (By default, a space will also be applied, see ``separate_value_from_prefix`` for more information)
        :param separate_value_from_prefix: (Default: True) Add a space between the prefix and value when ``True``.
        :param prefix_applies_to_all_elements: Applies the prefix to each element of the array (Array inputs only)
        :param shell_quote: Stops shell quotes from being applied in all circumstances, useful when joining multiple commands together.
        :param separator: The separator between each element of an array (defaults to ' ')
        :param localise_file: Ensures that the file(s) are localised into the execution directory.
        :param default: The default value to be applied if the input is not defined.
        :param doc: Documentation string for the ToolInput, this is used to generate the tool documentation and provide
        hints to the user.
        """
        super().__init__(
            value=None,
            prefix=prefix,
            position=position,
            separate_value_from_prefix=separate_value_from_prefix,
            doc=doc,
            shell_quote=shell_quote,
        )

        # if default is not None:
        #     input_type.optional = True

        if not Validators.validate_identifier(tag):
            raise Exception(
                f"The identifier '{tag}' was not validated because {Validators.reason_for_failure(tag)}"
            )

        self.tag: str = tag
        self.input_type: ParseableType = get_instantiated_type(input_type)
        self.default = default
        self.prefix_applies_to_all_elements = prefix_applies_to_all_elements
        self.separator = separator
        self.localise_file = localise_file

        # if isinstance(input_type, Array):
        #     if self.prefix_applies_to_all_elements is None and self.separator is None:
        # self.separator = " "

    def id(self):
        return self.tag


# This should really be a CommandToolOutput
class ToolOutput:
    def __init__(
        self,
        tag: str,
        output_type: ParseableType,
        glob: Optional[Union[Selector, str]] = None,
        doc: Optional[str] = None,
    ):
        """
        A ToolOutput instructs the the engine how to collect an output and how
        it may be referenced in a workflow.

        :param tag: The identifier of a output, must be unique in the inputs and outputs.
        :param output_type: The type of output that is being collected.
        :param glob: How to collect this output, can accept any :class:`janis.Selector`.
        :param doc: Documentation on what the output is, used to generate docs.
        """

        if not Validators.validate_identifier(tag):
            raise Exception(
                f"The identifier '{tag}' was invalid because {Validators.reason_for_failure(tag)}"
            )

        self.tag = tag
        self.output_type: ParseableType = get_instantiated_type(output_type)

        if not glob and not (
            isinstance(self.output_type, Stdout) or isinstance(self.output_type, Stderr)
        ):
            raise Exception(
                "ToolOutput expects a glob when the output type is not Stdout / Stderr"
            )

        self.glob = glob
        self.doc = doc

    def id(self):
        return self.tag


class CommandTool(Tool, ABC):
    """
    A CommandTool is an interface between Janis and a program to be executed.
    Simply put, a CommandTool has a name, a command, inputs, outputs and a container to run in.
    """

    def __init__(self, **connections):
        super().__init__(metadata_class=ToolMetadata, **connections)

    # Tool base
    @staticmethod
    @abstractmethod
    def tool() -> str:
        """
        Unique identifier of the tool
        :return:
        """
        pass

    @staticmethod
    @abstractmethod
    def base_command() -> Optional[Union[str, List[str]]]:
        """
        The command of the tool to execute, usually the tool name or path and not related to any inputs.
        This field will always come before any inputs or arguments, though it's possible to omit this
        field and the program will use the first ordered argument / position.
        :return: Optional[Union[str, List[str]]]
        """
        pass

    @abstractmethod
    def inputs(self) -> List[ToolInput]:
        """
        A list of named tool inputs that will be used to create the command line. See :class:`janis.ToolInput`
        for options on how to configure this command line binding.
        :return: List[janis.ToolInput]
        """
        pass

    def arguments(self) -> Optional[List[ToolArgument]]:
        """
        A list of arguments that will be used to create the command line. Although they are not directly
        addressable as inputs, it's possible to use use a :class:`janis.InputSelector` or
        :class:`janis.StringFormatter` in the value field. See :class:`janis.ToolArgument` for
        options on how to configure a this command line binding.
        :return: List[janis.ToolArgument]
        """
        return None

    @abstractmethod
    def outputs(self) -> List[ToolOutput]:
        """
        A list of named outputs of the tool. Each :class:`janis.ToolOutput` has a ``glob`` field that
        can be used to select the outputs, see its documentation for more information.
        :return:
        """
        pass

    def env_vars(self) -> Optional[Dict[str, Union[str, Selector]]]:
        return None

    # Tool versions

    @staticmethod
    @abstractmethod
    def container() -> str:
        """
        A link to an OCI compliant container accessible by your engine. Previously, docker().
        :return: str
        """
        pass

    @staticmethod
    @abstractmethod
    def version() -> str:
        """
        Version of the tool. Janis supports multiple versions of tools with the same ``.tool()`` value.
        The recommended format is `SemVer <https://semver.org/>`_, though you should reflect the tool version.
        :return: str
        """
        pass

    ## Other studd

    def id(self):
        return self.tool()

    @classmethod
    def __hash__(cls):
        return cls.tool()

    @classmethod
    def full_name(cls):
        if cls.version() is not None:
            return f"{cls.tool()}/{cls.version()}"
        return cls.tool()

    def memory(self, hints: Dict[str, Any]) -> Optional[float]:
        """
        These values are used to generate a separate runtime.json / runtime.yaml input
        that can be passed to the execution engine to fill in for the specified hints.

        These are now (2019-04-10) to be kept out of the workflow, to leave the workflow
        truly portable.

        This memory must be in GB!
        :param hints: Dict[Key: value] of hints
        :return: Optional[int]
        """
        return None

    def cpus(self, hints: Dict[str, Any]) -> Optional[int]:
        """
        These values are used to generate a separate runtime.json / runtime.yaml input
        that can be passed to the execution engine to fill in for the specified hints.

        These are now (2019-04-10) to be kept out of the workflow, to leave the workflow
        truly portable.

        The CPU must be a whole number. If your tool contains threads
        :return:
        """
        return None

    @classmethod
    def type(cls):
        return ToolTypes.CommandTool

    def translate(
        self,
        translation: SupportedTranslation,
        to_console=True,
        to_disk=False,
        export_path=None,
        with_docker=True,
        with_resource_overrides=False,
    ):
        import janis_core.translations

        return janis_core.translations.translate_tool(
            self,
            translation,
            to_console=to_console,
            to_disk=to_disk,
            export_path=export_path,
            with_docker=with_docker,
            with_resource_overrides=with_resource_overrides,
        )

    def tool_inputs(self) -> List[TInput]:
        return [TInput(t.id(), t.input_type, default=t.default) for t in self.inputs()]

    def tool_outputs(self) -> List[TOutput]:
        return [TOutput(t.id(), t.output_type) for t in self.outputs()]

    def all_input_keys(self):
        return super().all_input_keys() + [
            "runtime_memory",
            "runtime_cpu",
            "runtime_disks",
        ]

    def help(self):
        import inspect

        path = inspect.getfile(self.__class__)

        ins = sorted(
            self.inputs(), key=lambda i: i.position if i.position is not None else 0
        )
        args = ""
        if self.arguments():
            args = " " + " ".join(
                f"{(a.prefix if a.prefix is not None else '') + ' ' if (a.prefix is not None and a.separate_value_from_prefix) else ''}{a.value}"
                for a in self.arguments()
            )

        prefixes = " -" + "".join(
            i.prefix.replace("-", "").replace(" ", "")
            for i in ins
            if i.prefix is not None
        )

        metadata = self.metadata
        docker = self.container()

        base = (
            (
                self.base_command()
                if isinstance(self.base_command(), str)
                else " ".join(self.base_command())
            )
            if self.base_command()
            else ""
        )
        command = base + args + prefixes

        def input_format(t: ToolInput):
            prefix_with_space = ""
            if t.prefix is not None:
                prefix_with_space = (
                    (t.prefix + ": ")
                    if (t.separate_value_from_prefix is not False)
                    else t.prefix
                )
            return (
                f"\t\t{t.tag} ({prefix_with_space}{t.input_type.id()}{('=' + str(t.default)) if t.default is not None else ''})"
                f": {'' if t.doc is None else t.doc}"
            )

        output_format = (
            lambda t: f"\t\t{t.tag} ({t.output_type.id()}): {'' if t.doc is None else t.doc}"
        )

        requiredInputs = "\n".join(
            input_format(x) for x in ins if not x.input_type.optional
        )
        optionalInputs = "\n".join(
            input_format(x) for x in ins if x.input_type.optional
        )
        outputs = "\n".join(output_format(o) for o in self.outputs())

        return f"""
    Pipeline tool: {path} ({self.id()})
NAME
    {self.id()}

SYNOPSIS
    {command}

DOCKER
    {docker}

DOCUMENTATION URL
    {metadata.documentationUrl if metadata.documentationUrl else "No url provided"}

DESCRIPTION
    {metadata.documentation if metadata.documentation else "No documentation provided"}

INPUTS:
    REQUIRED:
{requiredInputs}

    OPTIONAL:
{optionalInputs}

OUTPUTS:
{outputs}
"""

    def generate_inputs_override(
        self, additional_inputs=None, with_resource_overrides=False, hints=None
    ):
        """
        Generate the overrides to be used with Janis. Although it may work with
        other
        :return:
        """
        d, ad = {}, additional_inputs or {}
        for i in self.inputs():
            if not i.input_type.optional or i.default or i.id() in ad:
                d[i] = ad.get(i.id(), i.default)

        if with_resource_overrides:
            cpus = self.cpus(hints) or 1
            mem = self.memory(hints)
            d.update(
                {
                    "runtime_memory": mem,
                    "runtime_cpu": cpus,
                    "runtime_disks": "local-disk 60 SSD",
                }
            )

        return d

    def wrapped_in_wf(self):
        from copy import copy
        from janis_core.workflow.workflow import WorkflowBuilder

        wf = WorkflowBuilder(self.id() + "Wf")
        inpmap = {}
        for i in self.inputs():

            if isinstance(i.input_type, Filename):
                intp = String(optional=True)
            else:
                intp = copy(i.input_type)
                if i.default:
                    intp.optional = True

            inpmap[i.id()] = wf.input(i.id(), intp)

        stp = wf.step(self.tool().lower(), self.__class__(**inpmap))

        for o in self.outputs():
            wf.output(o.id(), source=stp[o.id()])

        return wf



from keyword import iskeyword
from typing import Tuple, List

from janis_core.utils.logger import Logger


# https://stackoverflow.com/a/36331242/2860731
def variable_name_validator(x: str):
    return x.isidentifier() and not iskeyword(x)


def get_keywords_between_braces(
    text, argument_validator=variable_name_validator
) -> Tuple[set, int]:
    counter = 0
    highest_level = -1
    start_idx = None
    matches = set()
    rejected = set()
    skipped = set()

    for i in range(len(text)):

        char = text[i]
        if char == "{":
            counter += 1
            highest_level = max(highest_level, counter)
            if start_idx is None:
                start_idx = i
        elif char == "}" and counter > 0:
            counter -= 1

            if start_idx is not None and counter == 0:
                match = text[start_idx + 1 : i]
                if highest_level > 1:
                    skipped.add(match)
                elif argument_validator is not None and not argument_validator(match):
                    rejected.add(match)
                else:
                    matches.add(match)
                highest_level = -1
                start_idx = None
    extra: List[str] = []
    if len(matches) > 0:
        extra.append("matches=" + ",".join(matches))
    if len(rejected) > 0:
        extra.append("rejected=" + ",".join(rejected))
    if len(skipped) > 0:
        extra.append("skipped=" + ",".join(skipped))

    extrastr = ""
    if len(extra) > 0:
        extrastr = " (" + " | ".join(extra) + ")"

    Logger.log(f"Recognised {len(matches)} matches in '{text}'" + extrastr)

    return matches, counter



from setuptools import setup, find_packages

# Version information is found in the __init__ file of `janiscore/`
DESCRIPTION = "Contains classes and helpers to build a workflow, and provide options to convert to CWL / WDL"

######## SHOULDN'T NEED EDITS BELOW THIS LINE ########

with open("./README.md") as readme:
    long_description = readme.read()


#   Wondering why we have to get the version like this,
#   instead of simply from janis import __version__?
#
#   Well, well well, let me tell you...
#
#   This cost me days, but basically in `pip v19.0.0`,
#   they changed the way that modules are installed with PEP517.
#
#   Module installation is more isolated, and as this setup.py file called
#   for janis.__version__ AND we use a pyproject.toml file, pip / setuptools
#   gets confused and "From a PEP 517 perspective ..., it means it needs itself
#   to build. Which is of course a bit weird." [1] This means it is essentially
#   tries to import the project without resolving any of the setup.install_requires
#   dependencies. Installing the project with the `--no-use-pep517` flag works as expected.
#
#   I've elected to go with option (3) from [2]:
#
#   Links:
#       [1] https://github.com/pypa/pip/issues/6163#issuecomment-456738585
#       [2] https://packaging.python.org/guides/single-sourcing-package-version/
#       [3] https://github.com/pyinstaller/pyinstaller/issues/2730
#       [4] https://github.com/pypa/pip/issues/6163
#
vsn = {}
with open("./janis_core/__meta__.py") as fp:
    exec(fp.read(), vsn)
__version__ = vsn["__version__"]
githuburl = vsn["GITHUB_URL"]


setup(
    name="janis-pipelines.core",
    version=__version__,
    description=DESCRIPTION,
    url=githuburl,
    author="Michael Franklin, Evan Thomas, Mohammad Bhuyan",
    author_email="michael.franklin@petermac.org",
    license="GNU",
    keywords=["janis", "workflows", "pipelines", "cwl", "wdl", "bioinformatics"],
    packages=["janis_core"]
    + ["janis_core." + p for p in sorted(find_packages("./janis_core"))],
    install_requires=[
        "cwlgen >= 0.4.0",
        "illusional.wdlgen >= 0.2.5",
        "ruamel.yaml >= 0.12.4, <= 0.15.77",
        "tabulate",
        "path.py",
    ],
    # entry_points={"janis.extension": ["core=core"]},
    zip_safe=False,
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
)



"""
CWL

This is one of the more complicated classes, it takes the janis in-memory representation of a workflow,
and converts it into the equivalent CWL objects. Janis was built alongside testing for CWL, so a lot of
the concepts directly or pretty closely match. There are a few extra things that Janis has that need to
be mapped back.

This file is logically structured similar to the WDL equiv:

- Imports
- dump_cwl
- translate_workflow
- translate_tool (command tool)
- other translate methods
- selector helpers (InputSelector, WildcardSelector, CpuSelector, MemorySelector)
- helper methods
"""

## IMPORTS

import os
import re
from typing import List, Dict, Optional, Any, Tuple

import cwlgen
import ruamel.yaml

from janis_core.graph.steptaginput import full_lbl
from janis_core.tool.commandtool import CommandTool, ToolInput
from janis_core.tool.tool import Tool
from janis_core.translations.translationbase import TranslatorBase
from janis_core.types import (
    InputSelector,
    Selector,
    WildcardSelector,
    MemorySelector,
    CpuSelector,
    StringFormatter,
)
from janis_core.types.common_data_types import Stdout, Stderr, Array, File, Filename
from janis_core.utils import first_value
from janis_core.utils.logger import Logger
from janis_core.utils.metadata import WorkflowMetadata, ToolMetadata
from janis_core.translations.exportpath import ExportPathKeywords

# from janis_core.workflow.input import Input
from janis_core.workflow.workflow import StepNode

CWL_VERSION = "v1.0"


## TRANSLATION


class CwlTranslator(TranslatorBase):
    def __init__(self):
        super().__init__(name="cwl")
        ruamel.yaml.add_representer(
            cwlgen.utils.literal, cwlgen.utils.literal_presenter
        )

    @staticmethod
    def stringify_translated_workflow(wf):
        return ruamel.yaml.dump(wf.get_dict(), default_flow_style=False)

    @staticmethod
    def stringify_translated_tool(tool):
        return ruamel.yaml.dump(tool.get_dict(), default_flow_style=False)

    @staticmethod
    def stringify_translated_inputs(inputs):
        return ruamel.yaml.dump(inputs, default_flow_style=False)

    @staticmethod
    def validate_command_for(wfpath, inppath, tools_dir_path, tools_zip_path):
        return ["cwltool", "--validate", wfpath]

    @classmethod
    def translate_workflow(
        cls, wf, with_docker=True, with_resource_overrides=False, is_nested_tool=False
    ) -> Tuple[any, Dict[str, any]]:
        from janis_core.workflow.workflow import Workflow

        metadata = wf.metadata
        w = cwlgen.Workflow(
            wf.id(), wf.friendly_name(), metadata.documentation, cwl_version=CWL_VERSION
        )

        w.inputs: List[cwlgen.InputParameter] = [
            translate_input(i) for i in wf.input_nodes.values()
        ]

        resource_inputs = []
        if with_resource_overrides:
            resource_inputs = build_resource_override_maps_for_workflow(wf)
            w.inputs.extend(resource_inputs)

        w.steps: List[cwlgen.WorkflowStep] = []

        for s in wf.step_nodes.values():
            resource_overrides = {}
            for r in resource_inputs:
                if not r.id.startswith(s.id()):
                    continue

                resource_overrides[r.id[(len(s.id()) + 1) :]] = r.id
            w.steps.append(
                translate_step(
                    s,
                    is_nested_tool=is_nested_tool,
                    resource_overrides=resource_overrides,
                )
            )

        w.outputs = [translate_output_node(o) for o in wf.output_nodes.values()]

        w.requirements.append(cwlgen.InlineJavascriptRequirement())
        w.requirements.append(cwlgen.StepInputExpressionRequirement())

        if wf.has_scatter:
            w.requirements.append(cwlgen.ScatterFeatureRequirement())
        if wf.has_subworkflow:
            w.requirements.append(cwlgen.SubworkflowFeatureRequirement())
        if wf.has_multiple_inputs:
            w.requirements.append(cwlgen.MultipleInputFeatureRequirement())

        tools = {}
        tools_to_build: Dict[str, Tool] = {
            s.tool.id(): s.tool for s in wf.step_nodes.values()
        }
        for t in tools_to_build:
            tool: Tool = tools_to_build[t]
            if isinstance(tool, Workflow):
                wf_cwl, subtools = cls.translate_workflow(
                    tool,
                    is_nested_tool=True,
                    with_docker=with_docker,
                    with_resource_overrides=with_resource_overrides,
                )
                tools[tool.id()] = wf_cwl
                tools.update(subtools)
            elif isinstance(tool, CommandTool):
                tool_cwl = cls.translate_tool_internal(
                    tool,
                    with_docker=with_docker,
                    with_resource_overrides=with_resource_overrides,
                )
                tools[tool.id()] = tool_cwl
            else:
                raise Exception(f"Unknown tool type: '{type(tool)}'")

        return w, tools

    @classmethod
    def build_inputs_file(
        cls,
        workflow,
        recursive=False,
        merge_resources=False,
        hints=None,
        additional_inputs: Dict = None,
        max_cores=None,
        max_mem=None,
    ) -> Dict[str, any]:

        ad = additional_inputs or {}
        inp = {
            i.id(): i.datatype.cwl_input(ad.get(i.id()) or i.value or i.default)
            for i in workflow.input_nodes.values()
            if i.value or i.default or i.id() in ad
        }

        if merge_resources:
            for k, v in cls.build_resources_input(
                workflow, hints, max_cores, max_mem
            ).items():
                inp[k] = ad.get(k, v)

        return inp

    @classmethod
    def translate_workflow_to_all_in_one(
        cls, wf, with_resource_overrides=False, is_nested_tool=False
    ) -> cwlgen.Workflow:
        from janis_core.workflow.workflow import Workflow

        metadata = wf.metadata
        w = cwlgen.Workflow(
            wf.identifier,
            wf.friendly_name(),
            metadata.documentation,
            cwl_version=CWL_VERSION,
        )

        w.inputs: List[cwlgen.InputParameter] = [
            translate_input(i) for i in wf.input_nodes.values()
        ]

        resource_inputs = []
        if with_resource_overrides:
            resource_inputs = build_resource_override_maps_for_workflow(wf)
            w.inputs.extend(resource_inputs)

        w.steps: List[cwlgen.WorkflowStep] = []

        for s in wf.step_nodes.values():
            resource_overrides = {}
            for r in resource_inputs:
                if not r.id.startswith(s.id()):
                    continue

                resource_overrides[r.id[(len(s.id()) + 1) :]] = r.id

            w.steps.append(
                translate_step(
                    s,
                    is_nested_tool=is_nested_tool,
                    resource_overrides=resource_overrides,
                    use_run_ref=False,
                )
            )

        w.outputs = [translate_output_node(o) for o in wf._outputs]

        w.requirements.append(cwlgen.InlineJavascriptReq())
        w.requirements.append(cwlgen.StepInputExpressionRequirement())

        if wf.has_scatter:
            w.requirements.append(cwlgen.ScatterFeatureRequirement())
        if wf.has_subworkflow:
            w.requirements.append(cwlgen.SubworkflowFeatureRequirement())
        if wf.has_multiple_inputs:
            w.requirements.append(cwlgen.MultipleInputFeatureRequirement())

        return w

    @classmethod
    def translate_tool_internal(
        cls, tool: CommandTool, with_docker=True, with_resource_overrides=False
    ):
        metadata = tool.metadata if tool.metadata else ToolMetadata()
        stdouts = [
            o.outtype
            for o in tool.tool_outputs()
            if isinstance(o.outtype, Stdout) and o.outtype.stdoutname
        ]
        stderrs = [
            o.outtype
            for o in tool.tool_outputs()
            if isinstance(o.outtype, Stderr) and o.outtype.stderrname
        ]
        stdout = stdouts[0].stdoutname if len(stdouts) > 0 else None
        stderr = stderrs[0].stderrname if len(stderrs) > 0 else None

        if isinstance(stdout, InputSelector):
            stdout = translate_input_selector(stdout, code_environment=False)

        if isinstance(stderr, InputSelector):
            stderr = translate_input_selector(stderr, code_environment=False)

        tool_cwl = cwlgen.CommandLineTool(
            tool_id=tool.id(),
            base_command=tool.base_command(),
            label=tool.id(),
            doc=metadata.documentation,
            cwl_version=CWL_VERSION,
            stdin=None,
            stderr=stderr,
            stdout=stdout,
        )

        # if any(not i.shell_quote for i in tool.inputs()):
        tool_cwl.requirements.append(cwlgen.ShellCommandRequirement())

        tool_cwl.requirements.extend([cwlgen.InlineJavascriptRequirement()])

        envs = tool.env_vars()
        if envs:
            lls = [
                cwlgen.EnvVarRequirement.EnvironmentDef(
                    k,
                    get_input_value_from_potential_selector_or_generator(
                        value=v, code_environment=False, toolid=tool.id()
                    ),
                )
                for k, v in envs.items()
            ]
            tool_cwl.requirements.append(cwlgen.EnvVarRequirement(lls))

        inputs_that_require_localisation = [
            ti
            for ti in tool.inputs()
            if ti.localise_file
            and (
                issubclass(type(ti.input_type), File)
                or (issubclass(type(ti.input_type), Array))
                and issubclass(type(ti.input_type.subtype()), File)
            )
        ]
        if inputs_that_require_localisation:
            tool_cwl.requirements.append(
                cwlgen.InitialWorkDirRequirement(
                    [
                        "$(inputs.%s)" % ti.id()
                        for ti in inputs_that_require_localisation
                    ]
                )
            )

        if with_docker:
            tool_cwl.requirements.append(
                cwlgen.DockerRequirement(
                    docker_pull=tool.container(),
                    # docker_load=None,
                    # docker_file=None,
                    # docker_import=None,
                    # docker_image_id=None,
                    # docker_output_dir=None
                )
            )

        tool_cwl.inputs.extend(translate_tool_input(i) for i in tool.inputs())
        tool_cwl.outputs.extend(
            translate_tool_output(o, tool=tool.id()) for o in tool.outputs()
        )

        args = tool.arguments()
        if args:
            tool_cwl.arguments.extend(
                translate_tool_argument(a) for a in tool.arguments()
            )

        if with_resource_overrides:
            # work out whether (the tool of) s is a workflow or tool
            tool_cwl.inputs.extend(
                [
                    cwlgen.CommandInputParameter("runtime_memory", param_type="float?"),
                    cwlgen.CommandInputParameter("runtime_cpu", param_type="int?"),
                    # cwlgen.CommandInputParameter("runtime_disks", param_type="string?"),
                ]
            )

            tool_cwl.requirements.append(
                cwlgen.ResourceRequirement(
                    cores_min="$(inputs.runtime_cpu ? inputs.runtime_cpu : 1)",
                    ram_min="$(inputs.runtime_memory ? Math.floor(1024 * inputs.runtime_memory) : 4096)",
                )
            )

        return tool_cwl

    @staticmethod
    def workflow_filename(workflow):
        return workflow.id() + ".cwl"

    @staticmethod
    def inputs_filename(workflow):
        return workflow.id() + "-inp.yml"

    @staticmethod
    def tool_filename(tool):
        return (tool.id() if isinstance(tool, Tool) else str(tool)) + ".cwl"

    @staticmethod
    def resources_filename(workflow):
        return workflow.id() + "-resources.yml"


def translate_input(inp):
    """

    :param inp:
    :type inp: janis_core.workflow.workflow.InputNode
    :return:
    """
    return cwlgen.InputParameter(
        param_id=inp.id(),
        default=inp.default,
        secondary_files=inp.datatype.secondary_files(),
        param_format=None,
        streamable=None,
        doc=inp.doc,
        input_binding=None,
        param_type=inp.datatype.cwl_type(inp.default is not None),
    )


def translate_output_node(node):
    return translate_output(node, full_lbl(node.source[0], node.source[1]))


def translate_output(outp, source):
    ot = outp.datatype
    if isinstance(ot, Stdout):
        ot = ot.subtype or File()

    return cwlgen.WorkflowOutputParameter(
        param_id=outp.id(),
        output_source=source,
        secondary_files=outp.datatype.secondary_files(),
        param_format=None,
        streamable=None,
        doc=outp.doc,
        param_type=ot.cwl_type(),
        output_binding=None,
        linkMerge=None,
    )


def translate_tool_input(toolinput: ToolInput) -> cwlgen.CommandInputParameter:

    default, value_from = toolinput.default, None

    if isinstance(toolinput.input_type, Filename):
        default = toolinput.input_type.generated_filename()
        # value_from = get_input_value_from_potential_selector_or_generator(toolinput.input_type, code_environment=False, toolid=toolinput.id())
    elif is_selector(default):
        default = None
        value_from = get_input_value_from_potential_selector_or_generator(
            toolinput.default, code_environment=False, toolid=toolinput.id()
        )

    data_type = toolinput.input_type.cwl_type(default is not None)

    input_binding = cwlgen.CommandLineBinding(
        # load_contents=toolinput.load_contents,
        position=toolinput.position,
        prefix=toolinput.prefix,
        separate=toolinput.separate_value_from_prefix,
        item_separator=toolinput.separator,
        value_from=value_from,
        shell_quote=toolinput.shell_quote,
    )

    non_optional_dt_component = (
        [t for t in data_type if t != "null"][0]
        if isinstance(data_type, list)
        else data_type
    )

    # Binding array inputs onto the console
    # https://www.commonwl.org/user_guide/09-array-inputs/
    if isinstance(toolinput.input_type, Array) and isinstance(
        non_optional_dt_component, cwlgen.CommandInputArraySchema
    ):
        if toolinput.prefix_applies_to_all_elements:
            input_binding.prefix = None
            input_binding.separate = None
            nested_binding = cwlgen.CommandLineBinding(
                # load_contents=toolinput.load_contents,
                prefix=toolinput.prefix,
                separate=toolinput.separate_value_from_prefix,
                # item_separator=toolinput.item_separator,
                # value_from=toolinput.value_from,
                shell_quote=toolinput.shell_quote,
            )
            non_optional_dt_component.inputBinding = nested_binding

    return cwlgen.CommandInputParameter(
        param_id=toolinput.tag,
        label=toolinput.tag,
        secondary_files=toolinput.input_type.secondary_files(),
        # streamable=None,
        doc=toolinput.doc,
        input_binding=input_binding,
        default=default,
        param_type=data_type,
    )


def translate_tool_argument(argument):
    return cwlgen.CommandLineBinding(
        # load_contents=False,
        position=argument.position,
        prefix=argument.prefix,
        separate=argument.separate_value_from_prefix,
        # item_separator=None,
        value_from=get_input_value_from_potential_selector_or_generator(
            argument.value, code_environment=False
        ),
        shell_quote=argument.shell_quote,
    )


def translate_tool_output(output, **debugkwargs):
    return cwlgen.CommandOutputParameter(
        param_id=output.tag,
        label=output.tag,
        secondary_files=output.output_type.secondary_files(),
        # param_format=None,
        # streamable=None,
        doc=output.doc,
        output_binding=cwlgen.CommandOutputBinding(
            glob=translate_to_cwl_glob(
                output.glob, outputtag=output.tag, **debugkwargs
            ),
            # load_contents=False,
            # output_eval=None
        ),
        param_type=output.output_type.cwl_type(),
    )


def translate_step(
    step: StepNode,
    is_nested_tool=False,
    resource_overrides=Dict[str, str],
    use_run_ref=True,
):

    tool = step.tool
    if use_run_ref:
        run_ref = ("{tool}.cwl" if is_nested_tool else "tools/{tool}.cwl").format(
            tool=tool.id()
        )
    else:
        from janis_core.workflow.workflow import Workflow

        has_resources_overrides = len(resource_overrides) > 0
        if isinstance(tool, Workflow):
            run_ref = CwlTranslator.translate_workflow_to_all_in_one(
                tool, with_resource_overrides=has_resources_overrides
            )
        else:
            run_ref = CwlTranslator.translate_tool_internal(
                tool, True, with_resource_overrides=has_resources_overrides
            )

    cwlstep = cwlgen.WorkflowStep(
        step_id=step.id(),
        run=run_ref,
        # label=step.step.label,
        doc=step.doc,
        scatter=None,  # Filled by StepNode
        scatter_method=None,  # Filled by StepNode
    )

    cwlstep.out = [
        cwlgen.WorkflowStepOutput(output_id=o.tag) for o in step.tool.tool_outputs()
    ]

    ins = step.inputs()

    for k in ins:
        inp = ins[k]
        if k not in step.sources:
            if inp.intype.optional or inp.default:
                continue
            else:
                raise Exception(
                    f"Error when building connections for cwlstep '{step.id()}', "
                    f"could not find required connection: '{k}'"
                )

        edge = step.sources[k]
        ss = edge.slashed_source()
        link_merge = None

        if (
            ss is not None
            and not isinstance(ss, list)
            and isinstance(inp.intype, Array)
        ):
            start = edge.source().start
            outssval = start.outputs()
            source_type = (
                first_value(outssval)
                if len(outssval) == 1
                else outssval[edge.source().stag]
            ).outtype
            # has scattered = isinstance(start, StepNode) and start.scatter
            if not isinstance(source_type, Array) and not (
                isinstance(start, StepNode) and start.scatter
            ):
                ss = [ss]
                link_merge = "merge_nested"

        d = cwlgen.WorkflowStepInput(
            input_id=inp.tag,
            source=ss,
            link_merge=link_merge,  # this will need to change when edges have multiple source_map
            value_from=None,
        )

        cwlstep.inputs.append(d)

    for r in resource_overrides:
        cwlstep.inputs.append(
            cwlgen.WorkflowStepInput(input_id=r, source=resource_overrides[r])
        )

    if step.scatter:
        if len(step.scatter.fields) > 1:
            Logger.info(
                "Discovered more than one scatterable field on cwlstep '{step_id}', "
                "deciding scatterMethod to be '{method}".format(
                    step_id=step.id(), method=step.scatter.method
                )
            )
            cwlstep.scatterMethod = step.scatter.method.cwl()
        cwlstep.scatter = step.scatter.fields

    return cwlstep


## SELECTORS


def is_selector(selector):
    return issubclass(type(selector), Selector)


def get_input_value_from_potential_selector_or_generator(
    value, code_environment=True, **debugkwargs
):
    if value is None:
        return None
    if isinstance(value, str):
        return f'"{value}"' if code_environment else value
    elif isinstance(value, int) or isinstance(value, float):
        return value
    elif isinstance(value, Filename):
        # value.generated_filenamecwl() if code_environment else f"$({value.generated_filenamecwl()})"
        return (
            f'"{value.generated_filename()}"'
            if code_environment
            else value.generated_filename()
        )
    elif isinstance(value, StringFormatter):
        return translate_string_formatter(
            value, code_environment=code_environment, **debugkwargs
        )
    elif isinstance(value, InputSelector):
        return translate_input_selector(
            selector=value, code_environment=code_environment
        )
    elif isinstance(value, WildcardSelector):
        raise Exception(
            f"A wildcard selector cannot be used as an argument value for '{debugkwargs}'"
        )
    elif isinstance(value, CpuSelector):
        return translate_cpu_selector(value)
    elif isinstance(value, MemorySelector):
        return translate_memory_selector(value)
    elif callable(getattr(value, "cwl", None)):
        return value.cwl()

    raise Exception("Could not detect type %s to convert to input value" % type(value))


def translate_input_selector(selector: InputSelector, code_environment):
    if not selector.input_to_select:
        raise Exception("No input was selected for input selector: " + str(selector))

    basename_extra = ".basename" if selector.use_basename else ""
    base = f"inputs.{selector.input_to_select}{basename_extra}"
    return base if code_environment else f"$({base})"


def translate_string_formatter(
    selector: StringFormatter, code_environment=True, **debugkwargs
):

    escapedFormat = selector._format.replace("\\", "\\\\")

    if len(selector.kwargs) == 0:
        return escapedFormat

    kwargreplacements = [
        f".replace(/{re.escape('{' +k + '}')}/g, {get_input_value_from_potential_selector_or_generator(v, code_environment=True, **debugkwargs)})"
        for k, v in selector.kwargs.items()
    ]
    return f'$("{escapedFormat}"' + "".join(kwargreplacements) + ")"


def translate_to_cwl_glob(glob, **debugkwargs):
    if not glob:
        return None

    if not isinstance(glob, Selector):
        Logger.critical(
            "String globs are being phased out from tool output selections, please use the provided "
            "Selector (InputSelector or WildcardSelector) classes. " + str(debugkwargs)
        )
        return glob

    if isinstance(glob, InputSelector):
        return translate_input_selector(glob, code_environment=False)

    elif isinstance(glob, StringFormatter):
        return translate_string_formatter(glob)

    elif isinstance(glob, WildcardSelector):
        return glob.wildcard

    raise Exception("Unimplemented selector type: " + glob.__class__.__name__)


def translate_cpu_selector(selector: CpuSelector):
    return "$(inputs.runtime_cpu)"


def translate_memory_selector(selector: MemorySelector):
    return "$(Math.floor(inputs.runtime_memory))"


## OTHER HELPERS


def build_resource_override_maps_for_workflow(
    wf, prefix=None
) -> List[cwlgen.InputParameter]:
    from janis_core.workflow.workflow import Workflow

    # returns a list of key, value pairs
    inputs = []
    if not prefix:
        prefix = ""  # wf.id() + "."
    else:
        prefix += "_"

    for s in wf.step_nodes.values():
        tool: Tool = s.tool

        if isinstance(tool, CommandTool):
            tool_pre = prefix + s.id() + "_"
            inputs.extend(
                [
                    cwlgen.InputParameter(
                        tool_pre + "runtime_memory", param_type="float?"
                    ),
                    cwlgen.InputParameter(tool_pre + "runtime_cpu", param_type="int?"),
                    # cwlgen.InputParameter(tool_pre + "runtime_disks", param_type="string?"),
                ]
            )
        elif isinstance(tool, Workflow):
            tool_pre = prefix + s.id()
            inputs.extend(build_resource_override_maps_for_workflow(tool, tool_pre))

    return inputs



import re


class Validators:
    identifier_regex = r"[a-zA-Z][a-zA-Z0-9_]+\Z"
    compiled_identifier = re.compile(identifier_regex)

    extra_prohibited_keys = {
        "identifier",
        "tool",
        "scatter",
        "ignore_missing",
        "output",
        "input",
    }

    @staticmethod
    def validate_identifier(identifier) -> bool:
        if identifier in Validators.extra_prohibited_keys:
            return False
        a = Validators.compiled_identifier.match(identifier)
        return a is not None

    @staticmethod
    def reason_for_failure(identifier) -> str:
        if identifier in Validators.extra_prohibited_keys:
            return "it is a reserved keyword"
        a = Validators.compiled_identifier.match(identifier)
        if a is not None:
            return (
                f"it was not validated by '{Validators.identifier_regex}' "
                f"(must start with letters, and then only contain letters, numbers and an underscore)"
            )

        return "Undefined"



"""

It would be amazing to calculate the Levenshtein distance between two strings, so we could do a:
    "Did you mean: 'thisvariable'"

When you're connecting steps together.

Resources:
    - https://stackoverflow.com/a/5859823/2860731
    - http://stevehanov.ca/blog/index.php?id=114
"""



"""
    This file declares the 'entry_points' that Janis will look for in its operations, and
    that your tool can bind to in order to provide additional functionality to Janis.

    > Check this awesome tutorial about Snek Inc for more information about entry-points:
        https://amir.rachum.com/blog/2017/07/28/python-entry-points/

    This is how all of janis works, the unix and bioinformatics tools are registered under
    all three of 'janis.extension', 'janis.types' and 'janis.tools'. janis-runner is just
    bound under 'janis.extension' to provide 'janis.runner' and CLI functionality.
"""


"""
    All extensions registered under this entrypoint will be added as an export to 
    janis.{yourextension}. Additionally, they will be added to the python sys.path 
    so you can just call'import janis.{yourextension}'.
"""
EXTENSIONS = "janis.extension"


"""
    All extensions that want to declare types should register a 'janis.types' extension.
    The shed (`janis.shed`) will look up this entry point and traverse any child modules
    looking for data types. 
    
    Note: Users can still access your type through a direct Python import (preferred)

"""
DATATYPES = "janis.types"

"""
    All extensions that want to provide tools should register a 'janis.tools' extension.
    The shed (`janis.shed`) will look up this entry point and traverse any child modules
    looking for tools. 
    
    Note: Users can still access your tool through a direct Python import (preferred)
"""
TOOLS = "janis.tools"

"""
    Polished pipelines should be extended under the janis.pipelines class. This will allow
    the documentation to produce a separate page called "Pipelines" with these workflows.
    They are also rendered slightly differently.
"""
PIPELINES = "janis.pipelines"
